@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013,2016 MariaDB Corporation AB
+   Copyright (C) 2013,2019 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -89,15 +89,24 @@ MADB_DsnMap DsnMap[] = {
   {&DsnKeys[22], 4, txtSslCaPath,       260, 0},
   {&DsnKeys[23], 4, txtSslCipher,        32, 0},
   {&DsnKeys[24], 4, cbSslVerify,          0, 0},
-
+  {&DsnKeys[32], 4, cbTls11,              1, 0},
+  {&DsnKeys[32], 4, cbTls12,              2, 0},
+  {&DsnKeys[32], 4, cbTls13,              4, 0},
+  {&DsnKeys[33], 4, cbForceTls,           0, 0},
   {NULL, 0, 0, 0, 0}
 };
 
+#define CBGROUP_BIT(MapIdx)             (char)(DsnMap[MapIdx].MaxLength)
+#define IS_CB_CHECKED(MapIdx)           GetButtonState(DsnMap[MapIdx].Page, DsnMap[MapIdx].Item)
+#define CBGROUP_SETBIT(_Dsn, MapIdx)    *GET_FIELD_PTR(_Dsn, DsnMap[MapIdx].Key, char)|= CBGROUP_BIT(MapIdx)
+#define CBGROUP_RESETBIT(_Dsn, MapIdx)  *GET_FIELD_PTR(_Dsn, DsnMap[MapIdx].Key, char)&= ~CBGROUP_BIT(MapIdx)
+
 MADB_OptionsMap OptionsMap[]= {
   {1, rbPipe,                          MADB_OPT_FLAG_NAMED_PIPE},
   {2, ckReconnect,                     MADB_OPT_FLAG_AUTO_RECONNECT},
   {2, ckConnectPrompt,                 MADB_OPT_FLAG_NO_PROMPT},
   {2, ckCompressed,                    MADB_OPT_FLAG_COMPRESSED_PROTO},
+  {2, ckUseMycnf,                      MADB_OPT_FLAG_USE_CNF},
   {3, ckIgnoreSchema,                  MADB_OPT_FLAG_NO_SCHEMA},
   {3, ckIgnoreSpace,                   MADB_OPT_FLAG_IGNORE_SPACE},
   {3, ckMultiStmt,                     MADB_OPT_FLAG_MULTI_STATEMENTS},
@@ -148,6 +157,10 @@ my_bool SetDialogFields()
         SendDlgItemMessage(hwndTab[DsnMap[i].Page], DsnMap[i].Item, BM_SETCHECK,
                            Val ? BST_CHECKED : BST_UNCHECKED, 0);
       }
+      break;
+    case DSN_TYPE_CBOXGROUP:
+      SendDlgItemMessage(hwndTab[DsnMap[i].Page], DsnMap[i].Item, BM_SETCHECK,
+        (*GET_FIELD_PTR(Dsn, DsnMap[i].Key, char) & CBGROUP_BIT(i)) != '\0' ? BST_CHECKED : BST_UNCHECKED, 0);
     }
     i++;
   }
@@ -333,7 +346,17 @@ void GetDialogFields()
        *(int *)((char *)Dsn + DsnMap[i].Key->DsnOffset)= GetFieldIntVal(DsnMap[i].Page, DsnMap[i].Item);
        break;
     case DSN_TYPE_BOOL:
-      *(my_bool *)((char *)Dsn + DsnMap[i].Key->DsnOffset)=  GetButtonState(DsnMap[i].Page, DsnMap[i].Item);
+      *GET_FIELD_PTR(Dsn, DsnMap[i].Key, my_bool)=  IS_CB_CHECKED(i);
+      break;
+    case DSN_TYPE_CBOXGROUP:
+      if (IS_CB_CHECKED(i) != '\0')
+      {
+        CBGROUP_SETBIT(Dsn, i);
+      }
+      else
+      {
+        CBGROUP_RESETBIT(Dsn, i);
+      }
     }
     ++i;
   }
@@ -425,37 +448,68 @@ end:
 	  SQLFreeHandle(SQL_HANDLE_STMT, (SQLHANDLE)Stmt);
 }
 
-void MADB_WIN_TestDsn(my_bool ShowSuccess)
+
+static SQLRETURN TestDSN(MADB_Dsn *Dsn, SQLHANDLE *Conn)
 {
-  MYSQL *mysql= mysql_init(NULL);
   SQLHANDLE Connection= NULL;
-  SQLRETURN ret;
-  MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
-  
-  GetDialogFields();
+  SQLRETURN Result;
+  char * InitCommand= Dsn->InitCommand;
+
+  Dsn->InitCommand= NULL;
 
   SQLAllocHandle(SQL_HANDLE_DBC, Environment, (SQLHANDLE *)&Connection);
   assert(Connection != NULL);
-  
-  ret= ((MADB_Dbc *)Connection)->Methods->ConnectDB((MADB_Dbc *)Connection, Dsn);
+  Result= ((MADB_Dbc *)Connection)->Methods->ConnectDB((MADB_Dbc *)Connection, Dsn);
+
+  Dsn->InitCommand= InitCommand;
+
+  if (Conn != NULL)
+  {
+    *Conn= Connection;
+  }
+  else
+  {
+    SQLDisconnect(Connection);
+    SQLFreeHandle(SQL_HANDLE_DBC, Connection);
+  }
+
+  return Result;
+}
+
+
+void MADB_WIN_TestDsn(my_bool ShowSuccess)
+{
+  SQLHANDLE Connection;
+  SQLRETURN ret;
+  MADB_Dsn *Dsn= (MADB_Dsn *)GetWindowLongPtr(GetParent(hwndTab[0]), DWLP_USER);
+
+  GetDialogFields();
+
+  ret= TestDSN(Dsn, &Connection);
 
   if (ShowSuccess)
   {
     char Info[1024];
-	if (ret == SQL_SUCCESS)
+    if (SQL_SUCCEEDED(ret))
+    {
       _snprintf(Info, 1024, "Connection successfully established\n\nServer information: %s", mysql_get_server_info(((MADB_Dbc *)Connection)->mariadb));
-    MessageBox(hwndTab[CurrentPage],  (ret == SQL_SUCCESS) ? Info : 
-	         ((MADB_Dbc *)Connection)->Error.SqlErrorMsg, "Connection test",MB_ICONINFORMATION| MB_OK);
+      MessageBox(hwndTab[CurrentPage], Info, "Connection test", MB_ICONINFORMATION| MB_OK);
+    }
+    else
+    {
+      MessageBox(hwndTab[CurrentPage], ((MADB_Dbc *)Connection)->Error.SqlErrorMsg, "Connection test", MB_ICONINFORMATION| MB_OK);
+    }
   }
 
-  if (ret == SQL_SUCCESS)
+  if (SQL_SUCCEEDED(ret))
   {
     ConnectionOK= TRUE;
     DSN_Set_CharacterSets(Connection);
     DSN_Set_Database(Connection);
+
+    SQLDisconnect(Connection);
   }
 
-  SQLDisconnect(Connection);
   SQLFreeHandle(SQL_HANDLE_DBC, Connection);
 }
 
@@ -544,14 +598,14 @@ INT_PTR CALLBACK DialogDSNProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lPara
   case WM_CTLCOLORDLG:
     if (!hbrBg)
       hbrBg= CreateSolidBrush(RGB(255,255,255));
-      return (LONG)hbrBg;
+      return (INT_PTR)hbrBg;
     break;
    case WM_CTLCOLORSTATIC:
    {
      HDC hdcStatic = (HDC)wParam;
      SetTextColor(hdcStatic, RGB(0, 0, 0));
      SetBkMode(hdcStatic, TRANSPARENT);
-     return (LONG)hbrBg;
+     return (INT_PTR)hbrBg;
   }
   case WM_COMMAND:
     switch(LOWORD(wParam))
@@ -623,7 +677,7 @@ INT_PTR CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
   case WM_CTLCOLORDLG:
     if (!hbrBg)
       hbrBg= CreateSolidBrush(RGB(255,255,255));
-      return (LONG)hbrBg;
+      return (INT_PTR)hbrBg;
     break;
   case WM_CLOSE:
     if(MessageBox(hDlg, TEXT("Close the program?"), TEXT("Close"),
@@ -664,6 +718,7 @@ INT_PTR CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
   return FALSE;
 }
 
+
 BOOL APIENTRY DllMain( HMODULE hModule,
                        DWORD  ul_reason_for_call,
                        LPVOID lpReserved
@@ -673,6 +728,7 @@ BOOL APIENTRY DllMain( HMODULE hModule,
   return TRUE;
 }
 
+
 void CenterWindow(HWND hwndWindow)
 {
   RECT rectWindow;
@@ -691,7 +747,8 @@ void CenterWindow(HWND hwndWindow)
   
 }
 
-my_bool DSNDialog(HWND     hwndParent,
+
+BOOL DSNDialog(HWND     hwndParent,
                   WORD     fRequest,
                   LPCSTR   lpszDriver,
                   LPCSTR   lpszAttributes,
@@ -701,20 +758,18 @@ my_bool DSNDialog(HWND     hwndParent,
   BOOL    ret;
   char    *DsnName=  NULL;
   my_bool DsnExists= FALSE;
+  char    Delimiter= ';';
 
   if (Dsn->isPrompt < 0 || Dsn->isPrompt > MAODBC_PROMPT_REQUIRED)
   {
     Dsn->isPrompt= MAODBC_CONFIG;
   }
 
-  EffectiveDisabledPages=     DisabledPages[Dsn->isPrompt];
-  EffectiveDisabledControls=  DisabledControls[Dsn->isPrompt];
-
   if (lpszAttributes)
+  {
+    Delimiter= '\0';
     DsnName= strchr((char *)lpszAttributes, '=');
-  
-  if (lpszDriver)
-    MADB_DriverGet((char *)lpszDriver);
+  }
 
   if (DsnName)
   {
@@ -723,7 +778,9 @@ my_bool DSNDialog(HWND     hwndParent,
     if (!Dsn->isPrompt && !SQLValidDSN(DsnName))
     {
       if (hwndParent)
+      {
         MessageBox(hwndParent, "Validation of data source name failed", "Error", MB_ICONERROR | MB_OK);
+      }
       return FALSE;
     }
   }
@@ -739,43 +796,88 @@ my_bool DSNDialog(HWND     hwndParent,
 
   /* Even if DsnName invalid(in case of prompt) - we should not have problem */
   DsnExists= MADB_DSN_Exists(DsnName);
- 
-  InitCommonControls();
-
-  if (lpszDriver)
-    Dsn->Driver= _strdup(lpszDriver);
 
   SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &Environment);
   SQLSetEnvAttr(Environment, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, 0);
 
-  if (fRequest == ODBC_ADD_DSN)
+  switch (fRequest)
   {
-    if (DsnExists && hwndParent)
-    {
-      if (MessageBox(hwndParent, "Data source name already exists, do you want to replace it?", 
-                      "Question", MB_ICONQUESTION | MB_YESNO) != IDYES)
+    case ODBC_ADD_DSN:
+      if (DsnExists)
+      {
+        if (hwndParent)
+        {
+          if (MessageBox(hwndParent, "Data source name already exists, do you want to replace it?",
+            "Question", MB_ICONQUESTION | MB_YESNO) != IDYES)
+          {
+            return FALSE;
+          }
+        }
+      }
+      
+      MADB_DSN_SetDefaults(Dsn);
+      MADB_ParseConnString(Dsn, (char *)lpszAttributes, SQL_NTS, Delimiter);
+
+      /* Need to set driver after connstring parsing, and before saving */
+      if (lpszDriver)
+      {
+        Dsn->Driver= _strdup(lpszDriver);
+      }
+
+      if (SQL_SUCCEEDED(TestDSN(Dsn, NULL)))
+      {
+        return MADB_SaveDSN(Dsn);
+      }
+      else if (hwndParent == NULL)
+      {
         return FALSE;
-    }
-    Dsn->IsTcpIp= 1;
-  }
-  else
-  {
+      }
+
+      break;
+
+    case ODBC_CONFIG_DSN:
+
     /* i.e. not a prompt */
     if (Dsn->isPrompt == MAODBC_CONFIG && Dsn->SaveFile == NULL)
     {
       if (!DsnExists)
       {
-        MessageBox(0, "Data source name not found", "Error", MB_ICONERROR | MB_OK);
+        if (hwndParent != NULL)
+        {
+          MessageBox(0, "Data source name not found", "Error", MB_ICONERROR | MB_OK);
+        }
         return FALSE;
       }
-      else if (!MADB_ReadDSN(Dsn, (char *)lpszAttributes, TRUE))
+      else if (!MADB_ReadConnString(Dsn, (char *)lpszAttributes, SQL_NTS, Delimiter))
       {
         SQLPostInstallerError(ODBC_ERROR_INVALID_DSN, Dsn->ErrorMsg);
         return FALSE;
       }
+
+      if (hwndParent == NULL)
+      {
+        if (SQL_SUCCEEDED(TestDSN(Dsn, NULL)))
+        {
+          return MADB_SaveDSN(Dsn);
+        }
+        else if (hwndParent == NULL)
+        {
+          return FALSE;
+        }
+      }
     }
   }
 
+  InitCommonControls();
+
+  EffectiveDisabledPages=     DisabledPages[Dsn->isPrompt];
+  EffectiveDisabledControls=  DisabledControls[Dsn->isPrompt];
+
+  if (fRequest == ODBC_ADD_DSN && Dsn->isPrompt == MAODBC_CONFIG && Dsn->DSNName != NULL)
+  {
+    EffectiveDisabledControls= PromptDisabledControls;
+  }
+
   notCanceled= TRUE;
   hwndMain= CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_DIALOG1), 0, DialogProc, 0);
   SetWindowLongPtr(hwndMain, DWLP_USER, (LONG_PTR)Dsn);
@@ -784,9 +886,6 @@ my_bool DSNDialog(HWND     hwndParent,
   CurrentPage= -1;
   SetPage(hwndMain, 1);
 
-  Edit_SetReadOnly(GetDlgItem(hwndTab[0], txtDsnName), 
-                   (hwndParent && DsnName && fRequest == ODBC_ADD_DSN) ? TRUE : FALSE);
-
   SetDialogFields();
   CenterWindow(hwndMain);
   ShowWindow(hwndMain, SW_SHOW);
@@ -803,9 +902,15 @@ my_bool DSNDialog(HWND     hwndParent,
 
   SQLFreeHandle(SQL_HANDLE_ENV, Environment);
 
+  if (notCanceled && fRequest == ODBC_CONFIG_DSN && DsnName != NULL && strcmp(DsnName, Dsn->DSNName) != 0)
+  {
+    SQLRemoveDSNFromIni(DsnName);
+  }
+
   return notCanceled;
 }
 
+
 BOOL INSTAPI ConfigDSN(
      HWND     hwndParent,
      WORD     fRequest,
